(module com-nocandysw-linq
  (main show-examples))

(define (fold-left kons knil lst)
  "Call (KONS previous next) on each of LST in order, seeding with
KNIL."
  (match-case lst
    (() knil)
    ((?next . ?others)
     (fold-left kons (kons knil next) others))))

(define (xcons cdr car)
  "Reversed `cons'."
  (cons car cdr))

(define (partial f . first-args)
  "Answer F with FIRST-ARGS assumed."
  (lambda (. more-args)
    (apply f first-args more-args)))

(define (compose f . fs)
  "Like (f (fs... . args))."
  (match-case fs
    (() f)
    ((?fi) (lambda (. args) (f (apply fi args))))
    (?-
     (match-case (reverse (cons f fs))
       ((?n-arity . ?one-arities)
	(lambda (. args)
	  (fold-left (lambda (x f) (f x))
		     (apply n-arity args)
		     one-arities)))))))

;;; Thread each form as the first argument to its successor.
(define-syntax ->
  (syntax-rules ()
    ((_ form) form)
    ((_ form (outer-car outer-cdr ...) forms ...)
     (-> (outer-car form outer-cdr ...) forms ...))
    ((_ form fun forms ...)
     (-> (fun form) forms ...))))

(define (lazy-source lst)
  "Answer an even lazy list for LST."
  (if (list? lst)
      (delay (if (null? lst)
		 '()
		 (cons (car lst) (lazy-source (cdr lst)))))
      lst))				;be idempotent

(define (fold-lazy-left kons knil llst)
  "As with `fold-left', but for even lazy lists."
  (if (lazy-null? llst)
      knil
      (fold-lazy-left kons
		      (kons knil (lazy-car llst))
		      (lazy-cdr llst))))

(define (eager-sink llst)
  "Answer a normal list for even lazy LLST."
  (reverse! (fold-lazy-left xcons '() llst)))

(define (lazy-null? llst)
  "Answer whether LLST is empty."
  (null? (force llst)))

(define (lazy-car llst)
  "Answer the first of a lazy list."
  (car (force llst)))

(define (lazy-cdr llst)
  "Answer the remainder of a lazy list, as a lazy list."
  (cdr (force llst)))

(define (lazy-zip llst . llsts)
  "Answer all the first elements, then the second elements, etc."
  (apply lazy-map list llst llsts))

(define (lazy-map f lst . lsts)
  (delay (if (any? lazy-null? (cons lst lsts))
	     '()
	     (cons (apply f (lazy-car lst)
			  (map lazy-car lsts))
		   (apply lazy-map f (lazy-cdr lst)
			  (map lazy-cdr lsts))))))

(define (lazy-filter f lst)
  (delay (cond ((lazy-null? lst) '())
	       ((f (lazy-car lst))
		(cons (lazy-car lst)
		      (lazy-filter f (lazy-cdr lst))))
	       (else (force (lazy-filter f (lazy-cdr lst)))))))

;;; Provide the conventional invocation for remaining commands,
;;; handling the termination case.
(define-syntax sink-into
  (syntax-rules ()
    ((_ _ source) source)
    ((_ var source other-command other-args ...)
     (other-command var source other-args ...))))

(define-syntax from
  (syntax-rules (in)
    ((_ var in lst others ...)
     (sink-into var (lazy-source lst) others ...))))

(define-syntax where
  (syntax-rules ()
    ((_ var source pred others ...)
     (sink-into var (lazy-filter (lambda (var) pred) source)
		others ...))))

(define-syntax select
  (syntax-rules ()
    ((_ var source take others ...)
     (sink-into var (lazy-map (lambda (var) take) source)
		others ...))))

(define (show-examples _)
  "Print out some examples."
  (let-syntax ((explain (syntax-rules ()
			  ((_ shown ...)
			   (begin
			     (begin (print 'shown)
				    (display "=> ")
				    (print shown)) ...)))))
    ;; these are tests.
    (explain
     (eager-sink
      (from x in '(1 2 3) where (= x 2) select (+ x 42))))))

;;; Local variables:
;;; bee-body-indent: 2
;;; end:
